type category = [ `base | `godi | `apps | `conf ]
type relop = [`lt | `gt | `le | `ge | `eq ]
type options = [ `configure | `bmake | `opt | `htdoc ]
type dep = (category * string * (relop * string) option)
(*          category , name   , (relop , version)  *)
                    (* without category- *)

let string_of_relop = function
  | `lt -> "<"
  | `gt -> ">"
  | `le -> "<="
  | `ge -> ">="
  | `eq -> "=="

let string_of_category = function
  | `base -> "base"
  | `godi -> "godi"
  | `apps -> "apps"
  | `conf -> "conf"

(* should return something that looks like this:
   godi-ocaml>=3.06:../../godi/godi-ocaml
   TODO: clean this up with some let ins *)
let string_of_dep = function
    | (cat, name, None) ->
        (string_of_category cat) ^ "-" ^ name ^ "-[0-9]*:../../"
        ^ (string_of_category cat) ^ "/" ^ (string_of_category cat) ^ "-" ^ name
    | (cat, name, Some (relop, version)) ->
        (string_of_category cat) ^ "-" ^ name ^ (string_of_relop relop)
             ^ version ^ ":../../"
             ^ (string_of_category cat) ^ "/"
             ^ (string_of_category cat) ^ "-" ^ name

type t =
    {
      package: string; (* without category- *)
      version: string;
      revision: int;
      category: category;
      depends: dep list;
      build_depends: dep list;
      sources_site: string;
      sources_basename: string;
      sources_extension: string;
      sources_unpacksto: string; (* this will usually be sources_basename *)
      homepage: string;
      maintainer: string;
      short_desc: string;
      long_desc: string;
      (* everything from  here down wants to be optional *)
      options: options list;  (* this is optional, None = [] *)
      docfiles: string list;  (* this is optional, None = [] *)
      all_target: string option; (* defaults to `all' *)
      opt_target: string option; (* defaults to `opt' *)
      install_target: string option; (* defaults to `install' *)
    }

let sample =
    {
      package = "xml-light";
      version = "1.0";
      revision = 1;
      category = `godi;
      depends = [(`godi, "ocaml", None)];
      build_depends = [];
      sources_site = "http://www.xmllight.org/src/";
      sources_basename = "xml-light";
      sources_extension = ".tar.gz";
      sources_unpacksto = "xml-light";
      homepage = "http://www.xmllight.org/";
      maintainer = "ogunden@phauna.org";
      short_desc = "dame that is a sweet short desc";
      long_desc = "dame that is a sweet long desc";
      options = [`htdoc; `opt];
      docfiles = ["README"];
      all_target = None;
      opt_target = None;
      install_target = None;
    }

let descr c = c.long_desc

(* val makefile : t -> string *)
(* the entire makefile for the given control information as a giant string *)
(* TODO: use Format(?) *)
let makefile control =
  let lines = [
    "
# This file was automatically generated by GODOR
.include \"../../mk/bsd.prefs.mk\"
.include \"../../mk/godi.pkg.mk\"
";
    "VERSION=        " ^ control.version;
    "PKGNAME=        " ^ (string_of_category control.category)
                       ^ "-" ^ control.package ^ "-${VERSION}";
    "PKGREVISION=    " ^ (string_of_int control.revision);
    "DISTNAME=       " ^ control.sources_unpacksto;
    "DISTFILES=      " ^ control.sources_basename ^ control.sources_extension;
    "CATEGORIES=     " ^ string_of_category control.category;
    "MASTER_SITES=   " ^ control.sources_site ^ "/";
    "MAINTAINER=     " ^ control.maintainer;
    "HOMEPAGE=       " ^ control.homepage;
    "COMMENT=        " ^ control.short_desc;
    "";
    "AUTOGENERATE_PLIST = yes";
    (*"PKG := ${PKGNAME:S/-${VERSION}//}";*)
    "PKG  =               " ^ (string_of_category control.category)
                            ^ "-" ^ control.package;
    "MAKE_FLAGS=     PREFIX=${PREFIX}";
    (let deps = List.map (fun d -> "DEPENDS+=       " ^ (string_of_dep d))
                         control.depends in
    String.concat "\n" deps);
    (let bdeps = List.map (fun d -> "BUILD_DEPENDS+= " ^ (string_of_dep d))
                          control.build_depends in
    String.concat "\n" bdeps);
    "PATH:=          ${LOCALBASE}/bin:${PATH}";
    (if List.mem `configure control.options then
      "HAS_CONFIGURE = yes"
    else
      "#HAS_CONFIGURE = yes");
    (if List.mem `bmake control.options then
      "#USE_GMAKE = yes"
    else
      "USE_GMAKE = yes");
    "
pre-configure-copy:
.	if exists(files)
	    cd files && ${PAX} -rw -pp . ${WRKSRC}
.	endif

pre-configure: pre-configure-copy

pre-install-mkdirs:
.	for d in bin lib/ocaml/pkg-lib doc share man etc info sbin include
	    ${_PKG_SILENT}${_PKG_DEBUG}mkdir -p ${PREFIX}/${d}
.	endfor
.	for n in 1 2 3 4 5 6 7 8 9
	    ${_PKG_SILENT}${_PKG_DEBUG}mkdir -p ${PREFIX}/man/man${n}
.	endfor

pre-install: pre-install-mkdirs
";

    "ALL_TARGET=     all";
    ".if ${GODI_HAVE_OCAMLOPT} == \"yes\"";
    (if List.mem `opt control.options then
        "ALL_TARGET+=    opt"
    else
        "#ALL_TARGET+=    opt");
    ".endif";

    (match control.docfiles with
      [] -> ""
    | docs -> "
post-install:
	mkdir -p ${PREFIX}/doc/${PKG}
.	for DOC in " ^ (String.concat " " docs) ^ "
	    install -m 0644 ${WRKSRC}/${DOC} ${PREFIX}/doc/${PKG}
.	endfor");

    ".include \"../../mk/bsd.pkg.mk\"";
    "";
]


    in

    String.concat "\n" lines

let file_exists f =
    try
        Unix.access f [Unix.F_OK]; true
    with Unix.Unix_error (_,_,_) -> false

exception SysError of string

let write_package ~godi_localbase
                  ~control_file
                  ~control (* we have to pass both because of cyclic deps *)
                  ?filesdir
                  ~patches
                  ~delete () =

    (* sanity check *)
    if not (file_exists (godi_localbase ^ "/etc/godi.conf")) then
        raise (SysError (godi_localbase
                      ^ " does not look like a godi installation."));

    (* gather information *)
    let cat_s = string_of_category control.category in
    let base_dir =
        godi_localbase ^ "/build/" ^ cat_s ^ "/" ^ cat_s ^ "-" ^ control.package
    in
    let files_dir = base_dir ^ "/files" in
    let patches_dir = base_dir ^ "/patches" in
    let bmake = godi_localbase ^ "/bin/bmake" in

    (* test to see if the directory already exists *)
    (if file_exists base_dir then
        if delete then
          (match Sys.command ("rm -rf " ^ base_dir) with
            0 -> ()
          | n -> raise (SysError ("Unable to recursively delete " ^ base_dir)))
        else
          raise (SysError ("Can't write to " ^ base_dir ^ ": file exists.")));

    (* construct the package directory structure *)
    Unix.mkdir base_dir 0o755;
    Unix.mkdir files_dir 0o755;
    Unix.mkdir patches_dir 0o755;

    (* generate yon Makefile *)
    let mfchan = open_out (base_dir ^ "/Makefile") in
    output_string mfchan (makefile control);
    close_out mfchan;

    (* generate yon DESCR *)
    let descrchan = open_out (base_dir ^ "/DESCR") in
    output_string descrchan (descr control);
    close_out descrchan;

    (* dump in any patches *)
    List.iter
        (fun patchfile ->
            match Sys.command ("cp " ^ patchfile ^ " " ^ patches_dir) with
              0 -> ()
            | n -> raise (SysError
                   ("Unable to copy " ^ patchfile ^ " to " ^ patches_dir)))
        patches;

    (* dump in any files *)
    (match filesdir with
      None -> prerr_endline "No filesdir supplied, skipping"; ()
    | Some filesdir ->
        (match Sys.command ("cp -R " ^ filesdir ^ "/* " ^ files_dir) with
            0 -> ()
          | n -> raise (SysError ("Unable to recursively copy "
                                    ^ filesdir ^ "/* to " ^ files_dir))));

    (* generate yon checksums *)
    print_endline("Generating checksums..");
    if Sys.command ("cd " ^ base_dir ^ " && bmake makesum") <> 0 then
        raise (SysError ("Error generating checksum."));
    if Sys.command ("cd " ^ base_dir ^ " && bmake makepatchsum") <> 0 then
        raise (SysError ("Error generating patch checksums."));

    (* copy in the control file *)
    if Sys.command ("cp " ^ control_file ^ " " ^ base_dir) <> 0 then
        raise (SysError
                ("Unable to copy " ^ control_file ^ " to " ^ base_dir));

    print_endline ("Package written to " ^ base_dir ^ ".")
