{
open Controlparser

let comment_start_ps = ref []

let enter_comment lexbuf =
    let (start_p, end_p) =
        (Lexing.lexeme_start_p lexbuf, Lexing.lexeme_end_p lexbuf)
    in
    comment_start_ps := (start_p, end_p) :: !comment_start_ps

let leave_comment () =
    match !comment_start_ps with
        | [] -> assert false
        | _::ps -> comment_start_ps := ps

type lex_line =
    | Package | Version | Revision | Depends | Build_Depends | Sources
    | Homepage | Maintainer | Description | Options | Docfiles | Nothing

type lex_state = Lhs of lex_line | Rhs of lex_line

let lex_state = ref (Lhs Package)
let reset () = lex_state := Lhs Package

(* the order of the elements of this list is the order that we impose on
   the control file *)
let lhs_strings = [
        (* string, lex_line, token *)
        "Package:", Package, PACKAGE;
        "Version:", Version, VERSION;
        "Revision:", Revision, REVISION;
        "Depends:", Depends, DEPENDS;
        "Build-Depends:", Build_Depends, BUILD_DEPENDS;
        "Sources:", Sources, SOURCES;
        "Homepage:", Homepage, HOMEPAGE;
        "Maintainer:", Maintainer, MAINTAINER;
        "Description:", Description, DESCRIPTION;
        "Options:", Options, OPTIONS;
        "Docfiles:", Docfiles, DOCFILES
    ]

(* uses the lhs_strings list to determine what line comes next *)
let next_line line =
    let rec aux lhs_strings =
        match lhs_strings  with
          (_, l, _)::(_, l', _)::_ when l = line -> l'
        | _::rest -> aux rest
        | [] -> Nothing
    in aux lhs_strings

let string_of_lex_line line =
    (* three-way assoc mapping from the second to the first *)
    let assoc3_2_1 b abcs =
        let (a, _, _) = List.find (fun (a', b', c') -> b = b') abcs in a
    in
    try
        assoc3_2_1 line lhs_strings
    with
        | Not_found -> assert false

let lex_line_of_string s =
    let assoc3_1_2 a abcs =
        let (_, b, _) = List.find (fun (a', b', c') -> a = a') abcs in b
    in
    assoc3_1_2 s lhs_strings

let token_of_string s =
    let assoc3_1_3 a abcs =
        let (_, _, c) = List.find (fun (a', b', c') -> a = a') abcs in c
    in
    assoc3_1_3 s lhs_strings

(*** error reporting ***)
let newline lexbuf =
    let curr_p = lexbuf.Lexing.lex_curr_p in
    lexbuf.Lexing.lex_curr_p <-
        { curr_p with
          Lexing.pos_lnum = curr_p.Lexing.pos_lnum + 1;
          Lexing.pos_bol = curr_p.Lexing.pos_cnum }

let error lexbuf s =
    let pos1 = Lexing.lexeme_start_p lexbuf in
    let pos2 = Lexing.lexeme_end_p lexbuf in
    raise (Syntax.Error (pos1, pos2, s))

let invalid_char lexbuf lexing =
    error lexbuf ("Invalid character '" ^ Lexing.lexeme lexbuf
                 ^ "' while lexing " ^ lexing)

let check_equiv lexbuf s l =
    if (try lex_line_of_string s = l with Not_found -> false) then
        ()
    else
       error lexbuf ("Expected '" ^ string_of_lex_line l ^ "', got '" ^ s ^ "'")

let advance_state lexbuf string_seen =
    match string_seen, !lex_state with
        | (None, Rhs l) -> lex_state := Lhs (next_line l)
        | (None, Lhs l) -> ()
        | (Some s, Lhs l) -> check_equiv lexbuf s l; lex_state := Rhs l
        | (Some _, Rhs _) -> assert false
          (* when in an Rhs state, we only call advance_state with `None' *)

let lex_category lexbuf cat_str =
    match cat_str with
        | "apps" -> `apps
        | "godi" -> `godi
        | "conf" -> `conf
        | "base" -> `base
        | _ -> error lexbuf ("Invalid category, '" ^ cat_str
                            ^ "' in package name")
}

(*** regular expressions ***)
let digit = ['0'-'9']
(* must not contain dashes, according to pkgsrc documentation *)
(* XXX more documentation?  what *exactly* should we be matching with these? *)
let version_num = digit ['0'-'9' 'a'-'z' 'A'-'Z' '_' '.'] *
(* should these be extended?  should they be complementized? *)
(* let ident = ['a'-'z' 'A'-'Z' '-' '0'-'9' '_' '.'] + *)
let pkgname_cat = ['a'-'z' 'A'-'Z' '0'-'9' '_']+
let pkgname_name = ['a'-'z' 'A'-'Z' '0'-'9' '_' '-']+
let pkgname = (pkgname_cat as category) '-' (pkgname_name as name)
let pathname = ['a'-'z' 'A'-'Z' '-' '0'-'9' '_' '.' '/'] +
let url = ['a'-'z' 'A'-'Z' '-' '0'-'9' '_' '.' '/' ':' '~' '%' '+' '&' '#'] +

(* enormous hack to ensure that empty short descriptions don't cause the
 * lex_state to be advanced into Lhs Options mode *)
rule token = parse
    | [' ' '\t']        { token lexbuf }
    | ""                { match !lex_state with
                            | Rhs Description -> description lexbuf
                            | _ -> token' lexbuf }

and token' = parse
    | '\n'              { advance_state lexbuf None;
                          newline lexbuf;
                          token lexbuf }
    | [' ' '\t']        { token lexbuf }
    | "(*"              { enter_comment lexbuf; comment lexbuf; token lexbuf }
    | ""                { match !lex_state with
                            | Lhs Nothing -> error lexbuf ("Invalid trailing"
                                                          ^ " characters.")
                            | Lhs line -> lhs line lexbuf
                            | Rhs Package -> package lexbuf
                            | Rhs Version -> version lexbuf
                            | Rhs Revision -> revision lexbuf
                            | Rhs Depends -> depends lexbuf
                            | Rhs Build_Depends -> depends lexbuf
                            | Rhs Sources -> sources lexbuf
                            | Rhs Homepage -> homepage lexbuf
                            | Rhs Maintainer -> maintainer lexbuf
                            | Rhs Description -> description lexbuf
                            | Rhs Options -> options lexbuf
                            | Rhs Docfiles -> docfiles lexbuf
                            | Rhs Nothing -> assert false (* never happenz *) }
    | eof               { EOF }

and lhs line = parse
    (*
    | ("Package:" | "Version:" | "Revision:" | "Depends:" | "Build-Depends:"
      | "Sources:" | "Homepage:" | "Maintainer:" | "Description:" | "Options:"
      | "Docfiles:")
    *)
    | [^' ' '\t' '\n']* { let lexeme = Lexing.lexeme lexbuf in
                          advance_state lexbuf (Some lexeme);
                          token_of_string lexeme }

and package = parse
    | pkgname       { PKGNAME (lex_category lexbuf category, name) }
    | pkgname_cat   { error lexbuf ("Expected '-' in package name '"
                                   ^ Lexing.lexeme lexbuf ^ "'") }
    | _             { invalid_char lexbuf "package name" }

and version = parse
    | version_num   { STRING (Lexing.lexeme lexbuf) }
    | _             { invalid_char lexbuf "version number" }

and revision = parse
    | digit+        { INT (int_of_string (Lexing.lexeme lexbuf)) }
    | _             { invalid_char lexbuf "revision number" }

(* XXX here, and above, make pkgname_cat error message more informative:
 *  - if followed by whitespace, expected a dash.
 *  - otherwise, invalid character
 *)
and depends = parse
    | pkgname       { PKGNAME (lex_category lexbuf category, name) }
    | pkgname_cat   { error lexbuf ("Expected '-' in package name '"
                                   ^ Lexing.lexeme lexbuf ^ "'") }
    | version_num   { STRING (Lexing.lexeme lexbuf) }
    | '('           { LPAREN }
    | ')'           { RPAREN }
    | "=="          { RELOP `eq }
    | ">="          { RELOP `ge }
    | "<="          { RELOP `le }
    | '>'           { RELOP `gt }
    | '<'           { RELOP `lt }
    | ','           { COMMA }
    | _             { invalid_char lexbuf "dependencies" }

and sources = parse
    | url           { STRING (Lexing.lexeme lexbuf) }
    | _             { invalid_char lexbuf "sources url" }

and homepage = parse
    | url           { STRING (Lexing.lexeme lexbuf) }
    | _             { invalid_char lexbuf "homepage url" }

and maintainer = parse
    | [^'\n']*      { STRING (Lexing.lexeme lexbuf) }

and description = parse
    | _ * "\n.\n"   { advance_state lexbuf None; (* since we stole the last newline *)
                      let s = Lexing.lexeme lexbuf in
                      String.iter (function '\n' -> newline lexbuf | _ -> ()) s;
                      let split s c =
                        (* splits into (s, "") if c not in s *)
                        let len = String.length s in
                        try
                            let idx = String.index s c in
                            (String.sub s 0 idx,
                             String.sub s (idx + 1) (len - idx - 1))
                        with
                            Not_found -> (s, "")
                      in
                      let descrs = String.sub s 0 (String.length s - 2) in
                      let (short, long) = split descrs '\n' in
                      STRINGPAIR (short, long) }
    | ""            { error lexbuf "Unterminated description, expected '.' on a line by itself" }

and options = parse
    | "configure"   { OPTION `configure }
    | "bmake"       { OPTION `bmake }
    | "opt"         { OPTION `opt }
    | "htdoc"       { OPTION `htdoc }
    | ','           { COMMA }
    | [^' ' '\t' '\n' ',']*
                    { error lexbuf ("Invalid option '" ^ Lexing.lexeme lexbuf
                                    ^ "', expected one of "
                                    ^ "'configure', 'bmake', 'opt', 'htdoc'") }

and docfiles = parse
    | pathname      { STRING (Lexing.lexeme lexbuf) }
    | ','           { COMMA }
    | _             { invalid_char lexbuf "docfiles" }

and comment = parse
    | "*)"          { leave_comment ();
                      if !comment_start_ps = [] then () else comment lexbuf }
    | "(*"          { enter_comment lexbuf; comment lexbuf }
    | '\n'          { newline lexbuf; comment lexbuf }
    | _             { comment lexbuf }
    | eof           { match !comment_start_ps with
                        | [] -> assert false
                        | (p1, p2)::_ ->
                            raise (Syntax.Error (p1, p2, "Unterminated comment")) }

{ (* the trailer *) }
